#include <SPI.h>
#include <SD.h>
#include "ESP8266WiFi.h"
#include <SoftwareSerial.h>

WiFiClient client;

#define RXpin 4
#define TXpin 5
SoftwareSerial nss(RXpin, TXpin);
 
#define GPS nss

const unsigned char UBLOX_INIT[] PROGMEM = {
  // Disable NMEA
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x24, // GxGGA off
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x01,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x2B, // GxGLL off
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x02,0x00,0x00,0x00,0x00,0x00,0x01,0x02,0x32, // GxGSA off
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x03,0x00,0x00,0x00,0x00,0x00,0x01,0x03,0x39, // GxGSV off
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x04,0x00,0x00,0x00,0x00,0x00,0x01,0x04,0x40, // GxRMC off
  0xB5,0x62,0x06,0x01,0x08,0x00,0xF0,0x05,0x00,0x00,0x00,0x00,0x00,0x01,0x05,0x47, // GxVTG off

  // Disable UBX
  0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x17,0xDC, //NAV-PVT off
  0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x12,0xB9, //NAV-POSLLH off
  0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x13,0xC0, //NAV-STATUS off

  // Enable UBX
  0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x07,0x00,0x01,0x00,0x00,0x00,0x00,0x18,0xE1, //NAV-PVT on
  //0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x02,0x00,0x01,0x00,0x00,0x00,0x00,0x13,0xBE, //NAV-POSLLH on
  //0xB5,0x62,0x06,0x01,0x08,0x00,0x01,0x03,0x00,0x01,0x00,0x00,0x00,0x00,0x14,0xC5, //NAV-STATUS on

  // Rate
  //0xB5,0x62,0x06,0x08,0x06,0x00,0x64,0x00,0x01,0x00,0x01,0x00,0x7A,0x12, //(10Hz)
  0xB5,0x62,0x06,0x08,0x06,0x00,0xC8,0x00,0x01,0x00,0x01,0x00,0xDE,0x6A, //(5Hz)
  //0xB5,0x62,0x06,0x08,0x06,0x00,0xE8,0x03,0x01,0x00,0x01,0x00,0x01,0x39 //(1Hz)
};

const unsigned char UBX_HEADER[] = { 0xB5, 0x62 };

struct NAV_PVT {
  unsigned char cls;
  unsigned char id;
  unsigned short len;
  unsigned long iTOW;          // GPS time of week of the navigation epoch (ms)

  unsigned short year;         // Year (UTC) 
  unsigned char month;         // Month, range 1..12 (UTC)
  unsigned char day;           // Day of month, range 1..31 (UTC)
  unsigned char hour;          // Hour of day, range 0..23 (UTC)
  unsigned char minute;        // Minute of hour, range 0..59 (UTC)
  unsigned char second;        // Seconds of minute, range 0..60 (UTC)
  char valid;                  // Validity Flags (see graphic below)
  unsigned long tAcc;          // Time accuracy estimate (UTC) (ns)
  long nano;                   // Fraction of second, range -1e9 .. 1e9 (UTC) (ns)
  unsigned char fixType;       // GNSSfix Type, range 0..5
  char flags;                  // Fix Status Flags
  unsigned char reserved1;     // reserved
  unsigned char numSV;         // Number of satellites used in Nav Solution

  long lon;                    // Longitude (deg)
  long lat;                    // Latitude (deg)
  long height;                 // Height above Ellipsoid (mm)
  long hMSL;                   // Height above mean sea level (mm)
  unsigned long hAcc;          // Horizontal Accuracy Estimate (mm)
  unsigned long vAcc;          // Vertical Accuracy Estimate (mm)

  long velN;                   // NED north velocity (mm/s)
  long velE;                   // NED east velocity (mm/s)
  long velD;                   // NED down velocity (mm/s)
  long gSpeed;                 // Ground Speed (2-D) (mm/s)
  long heading;                // Heading of motion 2-D (deg)
  unsigned long sAcc;          // Speed Accuracy Estimate
  unsigned long headingAcc;    // Heading Accuracy Estimate
  unsigned short pDOP;         // Position dilution of precision
  short reserved2;             // Reserved
  unsigned long reserved3;     // Reserved
};

NAV_PVT pvt;
long lat = 0;
long lon = 0;

void calcChecksum(unsigned char* CK) {
  memset(CK, 0, 2);
  for (int i = 0; i < (int)sizeof(NAV_PVT); i++) {
    CK[0] += ((unsigned char*)(&pvt))[i];
    CK[1] += CK[0];
  }
}

bool processGPS() {
  static int fpos = 0;
  static unsigned char checksum[2];
  const int payloadSize = sizeof(NAV_PVT);

  while ( GPS.available() ) {
    byte c = GPS.read();
    if ( fpos < 2 ) {
      if ( c == UBX_HEADER[fpos] )
        fpos++;
      else
        fpos = 0;
    }
    else {      
      if ( (fpos-2) < payloadSize )
        ((unsigned char*)(&pvt))[fpos-2] = c;

      fpos++;

      if ( fpos == (payloadSize+2) ) {
        calcChecksum(checksum);
      }
      else if ( fpos == (payloadSize+3) ) {
        if ( c != checksum[0] )
          fpos = 0;
      }
      else if ( fpos == (payloadSize+4) ) {
        fpos = 0;
        if ( c == checksum[1] ) {
          return true;
        }
      }
      else if ( fpos > (payloadSize+4) ) {
        fpos = 0;
      }
    }
  }
  return false;
}

#ifndef DEGTORAD
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

struct geoloc {
  float lat;
  float lon;
  geoloc(float x = 0, float y = 0, int32_t a = 0) { lat = x; lon = y; }
};

// http://www.movable-type.co.uk/scripts/latlong.html
float geoDistance(struct geoloc &a, struct geoloc &b) { // meters
  const float R = 6371000; // km
  float p1 = a.lat * DEGTORAD;
  float p2 = b.lat * DEGTORAD;
  float dp = (b.lat-a.lat) * DEGTORAD;
  float dl = (b.lon-a.lon) * DEGTORAD;

  float x = sin(dp/2) * sin(dp/2) + cos(p1) * cos(p2) * sin(dl/2) * sin(dl/2);
  float y = 2 * atan2(sqrt(x), sqrt(1-x));

  return R * y;
}

geoloc lastLoggedLocation;
geoloc currentLocation;
 
String searchCameraSSID() {
 
  // Set WiFi to station mode and disconnect from an AP if it was previously connected
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  delay(100);

  // WiFi.scanNetworks will return the number of networks found
  int n = WiFi.scanNetworks();
  
  for (int i = 0; i < n; ++i) {
      if (WiFi.SSID(i).startsWith("YDXJ_")) 
        return WiFi.SSID(i);
      delay(10);
  }

  return "";
}
 
bool connectToCamera(String SSID) {
  char ssid[30];
  char *password = "1234567890";
  short retriesRemaining = 30;

  SSID.toCharArray(ssid, SSID.length() + 1);

  WiFi.begin(ssid, password);
  while ( WiFi.status() != WL_CONNECTED ) {
      if (retriesRemaining == 0)
          return false;
      delay(500);
      retriesRemaining--;
  }

  return true;
}

String requestToken() {
  
  client.print("{\"msg_id\":257,\"token\":0}\n\r");
  
  // Give the server time to respond, then read the entire reply
  delay(2000);
  String response;
  while (client.available()) {
    char character = client.read();
    response.concat(character);
  }

  // Search for token in the response
  
  String searchString = "\"param\":";
  int offset = response.indexOf(searchString);
  if (offset == -1) 
    return ""; // failure

  String token;
  for (int i = offset + searchString.length(); i < response.length(); ++i) {
    char c = response.charAt(i);
    if ((c == ',') || (c == ' ') || (c == '}') ) 
      break;
    else 
      token.concat(response.charAt(i));
  }

  return token;
}

void clearIncoming() {
  while (client.available())
    client.read();
}

File dataFile;
 
void shoot(String token) {

  clearIncoming();
  
  client.print("{\"msg_id\":769,\"token\":");
  client.print(token);
  client.print("}\n\r");

  if (dataFile) {
    dataFile.print(lat);
    dataFile.print(",");
    dataFile.println(lon);
    dataFile.flush();
  }

  clearIncoming();
}

bool haveGPSFix = false;
bool haveXiaomiConnection = false;
String token = "";

bool needPhoto = false;
unsigned long lastPhotoTime = 0;
int photoSeparationDistanceMeters = 10;
unsigned long minTimeBetweenPhotosMillis = 4000;

int gpsFixLedPin = 2; // this seems to be the onboard LED

const int sdChipSelect = 0;
char filename[16];

void setup() {
  GPS.begin(9600);

  SD.begin(sdChipSelect);

  delay(500);

  for (int i = 0; i < 1000; i++) {
    sprintf(filename, "data%d.txt", i);
    if ( ! SD.exists(filename) )
      break;
  }
  dataFile = SD.open(filename, FILE_WRITE);

  pinMode(gpsFixLedPin, OUTPUT);
  digitalWrite(gpsFixLedPin, LOW);
  
  currentLocation = geoloc(0,0);
  lastLoggedLocation = geoloc(0,0);

  // send configuration data in UBX protocol
  for(unsigned int i = 0; i < sizeof(UBLOX_INIT); i++) {                        
    GPS.write( pgm_read_byte(UBLOX_INIT+i) );
    delay(5); // simulating a 38400baud pace (or less), otherwise commands are not accepted by the device.
  }
}

unsigned long lastLedBlinkTime = 0;
unsigned long lastGPSPacketTime = 0;

void loop() {

  unsigned long now = millis();
  
  if ( now - lastGPSPacketTime > 1000 ) 
    haveGPSFix = false;

  int blinkTime = 1000;
  if ( haveXiaomiConnection && haveGPSFix )
    blinkTime = 100;
  else if ( haveXiaomiConnection )
    blinkTime = 500;
  else if ( haveGPSFix )
    blinkTime = 250;

  if ( now - lastLedBlinkTime > blinkTime ) {
    digitalWrite(gpsFixLedPin, !digitalRead(gpsFixLedPin));
    lastLedBlinkTime = now;
  }

  if ( processGPS() && pvt.fixType == 3 ) {

    lat = pvt.lat;
    lon = pvt.lon;
    currentLocation.lat = pvt.lat / 10000000.0f;
    currentLocation.lon = pvt.lon / 10000000.0f;
    
    if ( lastLoggedLocation.lat == 0 )
      lastLoggedLocation = currentLocation;
    
    float distanceToTarget = geoDistance( currentLocation, lastLoggedLocation );
    if ( distanceToTarget >= photoSeparationDistanceMeters ) 
      needPhoto = true;

    haveGPSFix = true;
    lastGPSPacketTime = millis();
  }
    
  if ( ! haveXiaomiConnection ) {
    String cameraSSID = searchCameraSSID();
    if (cameraSSID.length() == 0) 
      return;
      
    if (!connectToCamera(cameraSSID))
      return;

    if (!client.connect("192.168.42.1", 7878))
      return;

    token = requestToken();
    if (token.length() == 0) 
       return;

    haveXiaomiConnection = true;
  }
    
  if ( haveXiaomiConnection && needPhoto ) { 
    if ( millis() - lastPhotoTime > minTimeBetweenPhotosMillis ) {
      shoot(token);
      lastPhotoTime = millis();
      lastLoggedLocation = currentLocation;
      needPhoto = false;
    }
  }
}


