#include "ESP8266WiFi.h"
 
WiFiClient client;
 
String searchCameraSSID() {
 
  // Set WiFi to station mode and disconnect from an AP if it was previously connected
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  delay(100);

  // WiFi.scanNetworks will return the number of networks found
  int n = WiFi.scanNetworks();
  
  for (int i = 0; i < n; ++i) {
      if (WiFi.SSID(i).startsWith("YDXJ_")) 
        return WiFi.SSID(i);
      delay(10);
  }

  return "";
}
 
bool connectToCamera(String SSID) {
  char ssid[30];
  char *password = "1234567890";
  short retriesRemaining = 30;

  SSID.toCharArray(ssid, SSID.length() + 1);

  WiFi.begin(ssid, password);
  while ( WiFi.status() != WL_CONNECTED ) {
      if (retriesRemaining == 0)
          return false;
      delay(500);
      retriesRemaining--;
  }

  return true;
}

String requestToken() {
  
  client.print("{\"msg_id\":257,\"token\":0}\n\r");
  
  // Give the server time to respond, then read the entire reply
  delay(2000);
  String response;
  while (client.available()) {
    char character = client.read();
    response.concat(character);
    Serial.print(character);
  }

  Serial.println();

  // Search for token in the response
  
  String searchString = "\"param\":";
  int offset = response.indexOf(searchString);
  if (offset == -1) 
    return ""; // failure

  String token;
  for (int i = offset + searchString.length(); i < response.length(); ++i) {
    char c = response.charAt(i);
    if ((c == ',') || (c == ' ') || (c == '}') ) 
      break;
    else 
      token.concat(response.charAt(i));
  }

  return token;
}

void clearIncoming() {
  while (client.available())
    client.read();
}
 
void shoot(String token) {
  clearIncoming(); 
  
  client.print("{\"msg_id\":769,\"token\":");
  client.print(token);
  client.print("}\n\r");
  
  clearIncoming();
}


bool haveXiaomiConnection = false;
String token = "";

unsigned long lastPhotoTime = 0;
unsigned long minTimeBetweenPhotosMillis = 4000;

void setup() {
  Serial.begin(115200);
  Serial.println("Started");
}
 
void loop() {
    
  if ( ! haveXiaomiConnection ) {
    String cameraSSID = searchCameraSSID();
    if (cameraSSID.length() == 0) 
      return;

    Serial.println("Found network " + cameraSSID);
      
    if (!connectToCamera(cameraSSID))
      return;

    Serial.println("Connected to network");

    if (!client.connect("192.168.42.1", 7878))
      return;

    Serial.println("Connected to camera control server");

    token = requestToken();
    if (token.length() == 0) 
       return;

    Serial.println("Received token: " + token);

    haveXiaomiConnection = true;
  }
    
  if ( haveXiaomiConnection ) { 
    if ( millis() - lastPhotoTime > minTimeBetweenPhotosMillis ) {
      shoot(token);
      lastPhotoTime = millis();
    }
  }
}


